/*
 * $Id: nsCFHChannel.cpp,v 1.11 2004/01/22 00:39:56 schilz Exp $
 *
 * Copyright (c) 2003-2004  Dr. Thomas Schilz
 *
 * This file is part of MozPHP.
 *
 * MozPHP is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MozPHP is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MozPHP; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "nsCFHChannel.h"

#include "nsIStreamTransportService.h"
#include "nsIFileURL.h"
#include "nsIURL.h"
#include "nsIFile.h"
#include "nsIServiceManager.h"
#include "nsIMIMEService.h"
#include "nsMimeTypes.h"
#include "nsEventQueueUtils.h"
#include "nsXPIDLString.h"
#include "nsNetUtil.h"
#include "nsIURLParser.h"


#define LOG(x)  /* printf ("%s: %s\n", __FILE__, x); */

#define NOT_IMPLEMENTED(x) \
    LOG (x) \
    return NS_ERROR_NOT_IMPLEMENTED;

static NS_DEFINE_CID(kStreamTransportServiceCID, NS_STREAMTRANSPORTSERVICE_CID);

//-----------------------------------------------------------------------------

nsCFHChannel::nsCFHChannel()
	: mContentLength(-1),
      mLoadFlags(nsIFileChannel::LOAD_NORMAL),
      mStatus(NS_OK)
{
	NS_INIT_ISUPPORTS();
}

nsCFHChannel::~nsCFHChannel()
{
}


//-----------------------------------------------------------------------------
// nsCFHChannel methods
//-----------------------------------------------------------------------------

nsresult nsCFHChannel::Init (nsIURI* uri)
{
    mURI = uri;

    nsresult rv;
    mFileURL = do_QueryInterface (mURI, &rv);
    if (NS_FAILED(rv)) return rv;

    NS_ENSURE_TRUE (mFileURL, NS_ERROR_UNEXPECTED);

	return NS_OK;
}


nsresult nsCFHChannel::GetClonedFile (nsIFile **result)
{
    nsresult rv;

    nsCOMPtr<nsIFile> file;
    rv = GetFile (getter_AddRefs (file));
    if (NS_FAILED(rv)) return rv;

    return file->Clone(result);
}


nsresult nsCFHChannel::EnsureStream()
{
    NS_ENSURE_TRUE (mFileURL, NS_ERROR_NOT_INITIALIZED);

    nsCOMPtr<nsIFile> file;

    // don't assume nsIFile impl is threadsafe; pass a clone to the stream.
    nsresult rv = GetClonedFile (getter_AddRefs (file));

    if (NS_FAILED (rv)) return rv;

	PRBool is_dir;

    // we accept that this might result in a disk hit to stat the file
    rv = file->IsDirectory (&is_dir);
    if (NS_FAILED (rv)) {
        // canonicalize error message
        if (rv == NS_ERROR_FILE_TARGET_DOES_NOT_EXIST)
            rv = NS_ERROR_FILE_NOT_FOUND;
        return rv;
    }

    if (is_dir)
		return NS_ERROR_INVALID_ARG;	// we don't accept directories

    rv = NS_NewLocalFileInputStream (getter_AddRefs (mStream), file);
    if (NS_FAILED (rv)) return rv;

    // get content type from file extension

	/* <ugly_hack>
	 *
	 * We can't rely on getting the MIME type which was entered in the "Helper
	 * Applications" dialog, using GetTypeFromFile. After checking the builtin
	 * MIME types Mozilla first consults the OS to find a MIME type for a given
	 * extension. See nsExternalHelperAppService::GetTypeFromFile.
	 * 
	 * Under Windows the Python installer adds an entry "Content type = text/plain"
	 * for extension .py to the registry in HKEY_CLASSES_ROOT and this is then
	 * used by Mozilla in GetTypeFromFile.
	 *
	 * So we make sure here that for the extensions we want to handle in MozPHP
	 * and MozPython the desired MIME type is assigned to the channel's ContentType.
	 *
	 * </ugly_hack>
	 */
    nsXPIDLCString mimeType;

	nsCAutoString ext;
	rv = mFileURL->GetFileExtension (ext);
	if (NS_SUCCEEDED (rv)  &&  !ext.IsEmpty()) {
			 if (!strcmp (ext.get(), "py"))
				mimeType.Assign (NS_LITERAL_CSTRING ("text/python"));
		else if (!strcmp (ext.get(), "php"))
				mimeType.Assign (NS_LITERAL_CSTRING ("text/php"));
	}

    if (mimeType.IsEmpty()) {
		// Extension is neither .py nor .php
    	nsCOMPtr<nsIMIMEService> mime = do_GetService ("@mozilla.org/mime;1", &rv);
    	if (NS_SUCCEEDED (rv))
        	mime->GetTypeFromFile (file, getter_Copies (mimeType));
	}

    if (mimeType.IsEmpty()) {
        mContentType = NS_LITERAL_CSTRING (UNKNOWN_CONTENT_TYPE);
	}
    else
        mContentType = mimeType;

    // fixup content length
    if (mStream && (mContentLength < 0))
        mStream->Available ((PRUint32 *) &mContentLength);

    return NS_OK;

}


NS_METHOD nsCFHChannel::Create
	(nsISupports* aOuter, const nsIID& aIID, void* *aResult)
{
	if (aOuter)
		return NS_ERROR_NO_AGGREGATION;

	if (!aResult)
		return NS_ERROR_NULL_POINTER;

	*aResult = nsnull;

	nsCFHChannel* fc = new nsCFHChannel();

	if (fc == nsnull)
		return NS_ERROR_OUT_OF_MEMORY;

	NS_ADDREF(fc);
	nsresult rv = fc->QueryInterface(aIID, aResult);
	NS_RELEASE(fc);

	return rv;
}

//-----------------------------------------------------------------------------

NS_IMPL_THREADSAFE_ADDREF(nsCFHChannel)
NS_IMPL_THREADSAFE_RELEASE(nsCFHChannel)
NS_INTERFACE_MAP_BEGIN(nsCFHChannel)
	NS_INTERFACE_MAP_ENTRY(nsIFileChannel)
	NS_INTERFACE_MAP_ENTRY(nsIHttpChannel)
	NS_INTERFACE_MAP_ENTRY(nsIHttpChannelInternal)
	NS_INTERFACE_MAP_ENTRY(nsIStreamListener)
	NS_INTERFACE_MAP_ENTRY(nsITransportEventSink)
	NS_INTERFACE_MAP_ENTRY(nsIUploadChannel)
	NS_INTERFACE_MAP_ENTRY(nsIPostData)
	NS_INTERFACE_MAP_ENTRY_AMBIGUOUS(nsIChannel,nsIFileChannel)
	NS_INTERFACE_MAP_ENTRY_AMBIGUOUS(nsISupports,nsIFileChannel)
	NS_INTERFACE_MAP_ENTRY_AMBIGUOUS(nsIRequest,nsIFileChannel)
	NS_INTERFACE_MAP_ENTRY_AMBIGUOUS(nsIRequestObserver, nsIStreamListener)
NS_INTERFACE_MAP_END

//-----------------------------------------------------------------------------
// nsIRequest methods
//-----------------------------------------------------------------------------

NS_IMETHODIMP nsCFHChannel::GetName (nsACString &result)
{
    if (mRequest)
        return mRequest->GetName(result);
    return mURI->GetSpec(result);
}

NS_IMETHODIMP nsCFHChannel::IsPending (PRBool *result)
{
    if (mRequest)
        return mRequest->IsPending(result);
    *result = PR_FALSE;
    return NS_OK;
}

NS_IMETHODIMP nsCFHChannel::GetStatus (nsresult *status)
{
    *status = mStatus;
    return NS_OK;
}

NS_IMETHODIMP nsCFHChannel::Cancel (nsresult status)
{
    NS_ASSERTION(NS_FAILED(status), "shouldn't cancel with a success code");
    mStatus = status;
    if (mRequest)
        return mRequest->Cancel(status);
    return NS_OK;
}

NS_IMETHODIMP nsCFHChannel::Suspend()
{
    if (mRequest)
        return mRequest->Suspend();
    return NS_OK;
}

NS_IMETHODIMP nsCFHChannel::Resume()
{
    if (mRequest)
        return mRequest->Resume();
    return NS_OK;
}

//-----------------------------------------------------------------------------
// nsIChannel methods
//-----------------------------------------------------------------------------

NS_IMETHODIMP nsCFHChannel::GetOriginalURI (nsIURI* *aURI)
{
    *aURI = mOriginalURI ? mOriginalURI : mURI;
    NS_ADDREF(*aURI);
    return NS_OK;
}


NS_IMETHODIMP nsCFHChannel::SetOriginalURI (nsIURI* aURI)
{
    mOriginalURI = aURI;
    return NS_OK;
}


NS_IMETHODIMP nsCFHChannel::GetURI (nsIURI* *aURI)
{
    *aURI = mURI;
    NS_ADDREF(*aURI);
    return NS_OK;
}


NS_IMETHODIMP nsCFHChannel::GetLoadFlags (PRUint32 *aLoadFlags)
{
    *aLoadFlags = mLoadFlags;
    return NS_OK;
}


NS_IMETHODIMP nsCFHChannel::SetLoadFlags (PRUint32 aLoadFlags)
{
    mLoadFlags = aLoadFlags;
    return NS_OK;
}


NS_IMETHODIMP nsCFHChannel::GetContentType (nsACString &aContentType)
{
    aContentType = mContentType;
    return NS_OK;
}


NS_IMETHODIMP nsCFHChannel::SetContentType (const nsACString &aContentType)
{
    // only modifies mContentCharset if a charset is parsed
    NS_ParseContentType (aContentType, mContentType, mContentCharset);
    return NS_OK;
}


NS_IMETHODIMP nsCFHChannel::GetContentCharset (nsACString &aContentCharset)
{
    aContentCharset = mContentCharset;
    return NS_OK;
}


NS_IMETHODIMP nsCFHChannel::SetContentCharset (const nsACString &aContentCharset)
{
    mContentCharset = aContentCharset;
    return NS_OK;
}


NS_IMETHODIMP nsCFHChannel::GetContentLength (PRInt32 *aContentLength)
{
	*aContentLength = mContentLength;
	return NS_OK;
}


NS_IMETHODIMP nsCFHChannel::SetContentLength (PRInt32 aContentLength)
{
    NOT_IMPLEMENTED("nsCFHChannel::SetContentLength");
}


NS_IMETHODIMP nsCFHChannel::GetLoadGroup (nsILoadGroup* *aLoadGroup)
{
    *aLoadGroup = mLoadGroup;
    NS_IF_ADDREF (*aLoadGroup);
    return NS_OK;
}


NS_IMETHODIMP nsCFHChannel::SetLoadGroup (nsILoadGroup* aLoadGroup)
{
	mLoadGroup = aLoadGroup;
	return NS_OK;
}


NS_IMETHODIMP nsCFHChannel::GetOwner (nsISupports* *aOwner)
{
    *aOwner = mOwner.get();
    NS_IF_ADDREF (*aOwner);
    return NS_OK;
}


NS_IMETHODIMP nsCFHChannel::SetOwner (nsISupports* aOwner)
{
    mOwner = aOwner;
    return NS_OK;
}


NS_IMETHODIMP nsCFHChannel::GetNotificationCallbacks
	(nsIInterfaceRequestor* *aNotificationCallbacks)
{
	*aNotificationCallbacks = mCallbacks.get();
	NS_IF_ADDREF (*aNotificationCallbacks);
	return NS_OK;
}


NS_IMETHODIMP nsCFHChannel::SetNotificationCallbacks
	(nsIInterfaceRequestor* aNotificationCallbacks)
{
	mCallbacks = aNotificationCallbacks;

	if (mCallbacks)
		mProgressSink = do_GetInterface (mCallbacks);

	return NS_OK;
}


NS_IMETHODIMP nsCFHChannel::GetSecurityInfo (nsISupports * *aSecurityInfo)
{
    *aSecurityInfo = nsnull;
    return NS_OK;
}


NS_IMETHODIMP nsCFHChannel::Open (nsIInputStream **result)
{
    NOT_IMPLEMENTED ("nsCFHChannel::Open");
}


NS_IMETHODIMP nsCFHChannel::AsyncOpen
	(nsIStreamListener *listener, nsISupports *ctx)
{
	NS_ENSURE_TRUE(!mRequest, NS_ERROR_IN_PROGRESS);

    nsresult rv;

    nsCOMPtr<nsIEventQueue> currentEventQ;

    rv = NS_GetCurrentEventQ (getter_AddRefs (currentEventQ));
    if (NS_FAILED (rv)) return rv;

    nsCOMPtr<nsIStreamTransportService> sts =
            do_GetService (kStreamTransportServiceCID, &rv);
    if (NS_FAILED (rv)) return rv;

    //
    // create file input stream
    //
    rv = EnsureStream();
    if (NS_FAILED(rv)) return rv;

    //
    // create asynchronous input stream wrapping file input stream.
    //
    nsCOMPtr<nsITransport> transport;
    rv = sts->CreateInputTransport (mStream, -1, -1, PR_TRUE,
                                    getter_AddRefs (transport));
    if (NS_FAILED(rv)) return rv;

    rv = transport->SetEventSink (this, currentEventQ);
    if (NS_FAILED (rv)) return rv;

    nsCOMPtr<nsIInputStream> asyncIn;
    rv = transport->OpenInputStream (0, 0, 0, getter_AddRefs (asyncIn));
    if (NS_FAILED (rv)) return rv;

    //
    // create input stream pump
    //
    nsCOMPtr<nsIInputStreamPump> pump;
    rv = NS_NewInputStreamPump (getter_AddRefs (pump), asyncIn);
    if (NS_FAILED (rv)) return rv;

    rv = pump->AsyncRead (this, nsnull);
    if (NS_FAILED (rv)) return rv;

    mRequest = pump;

    if (mLoadGroup)
        mLoadGroup->AddRequest (NS_STATIC_CAST (nsIFileChannel*, this), nsnull);

    mListener = listener;
    mListenerContext = ctx;
    return NS_OK;
}

//-----------------------------------------------------------------------------
// nsIRequestObserver methods:
//-----------------------------------------------------------------------------

NS_IMETHODIMP nsCFHChannel::OnStartRequest
	(nsIRequest* request, nsISupports* context)
{
    // unconditionally inherit the status of the file transport
    request->GetStatus (&mStatus);

    NS_ASSERTION (mListener, "No listener...");

    nsresult rv = NS_OK;
    if (mListener)
        rv = mListener->OnStartRequest
					(NS_STATIC_CAST (nsIFileChannel*, this), context);
    return rv;
}


NS_IMETHODIMP nsCFHChannel::OnStopRequest
	(nsIRequest* request, nsISupports* context, nsresult aStatus)
{
    if (NS_SUCCEEDED (mStatus))
        mStatus = aStatus;

    NS_ENSURE_TRUE (mListener, NS_ERROR_UNEXPECTED);

    mListener->OnStopRequest
		(NS_STATIC_CAST (nsIFileChannel*, this), context, mStatus);

    if (mLoadGroup)
        mLoadGroup->RemoveRequest
			(NS_STATIC_CAST (nsIFileChannel*, this), context, mStatus);

    mListener = 0;
    mRequest = 0;
	mStream = 0;

    return NS_OK;
}

//-----------------------------------------------------------------------------
// nsIStreamListener methods:
//-----------------------------------------------------------------------------

NS_IMETHODIMP nsCFHChannel::OnDataAvailable
	(nsIRequest* request, nsISupports* context,
	 nsIInputStream *aIStream, PRUint32 aSourceOffset, PRUint32 aLength)
{
    NS_ENSURE_TRUE (mListener, NS_ERROR_UNEXPECTED);

    return mListener->OnDataAvailable (NS_STATIC_CAST (nsIFileChannel*, this),
									context, aIStream, aSourceOffset, aLength);
}


//-----------------------------------------------------------------------------
// nsIFileChannel methods
//-----------------------------------------------------------------------------

NS_IMETHODIMP nsCFHChannel::GetFile (nsIFile* *result)
{
	/*
	 * Ugly HACK: We change the scheme to "file" temporarily to be able to
	 * create a nsILocalFile instance. With scheme "cfh" several checks
	 * fail for example in nsStandardURL.cpp.
	 * We must revert the scheme to "cfh" afterwards. Otherwise URLs
	 * relative to the one just loaded will get a "file" scheme.
	 */
	mURI->SetScheme (nsCAutoString ("file"));

	nsresult rv = mFileURL->GetFile (result);

	mURI->SetScheme (nsCAutoString (SCHEME));

    return rv;
}


//-----------------------------------------------------------------------------
// nsIUploadChannel methods
//-----------------------------------------------------------------------------

NS_IMETHODIMP nsCFHChannel::SetUploadStream
	(nsIInputStream *aStream, const nsACString &aContentType,
	 PRInt32 aContentLength)
{
	NS_ENSURE_TRUE (!mRequest, NS_ERROR_IN_PROGRESS);

	PRUint32 upload_stream_length;

    if (aStream) {
    	if (aContentLength < 0) {
       	     nsresult rv = aStream->Available (&upload_stream_length);
       	     if (NS_FAILED(rv)) return rv;
    	}
	    else
        	upload_stream_length = aContentLength;

		char buf[256];
		unsigned long read_total = 0;
		mPostData.Assign ("");
		while (read_total < upload_stream_length) {
			PRUint32 bytes_read;
			nsresult rv = aStream->Read ((char*) buf, unsigned (255), &bytes_read);
			if (NS_FAILED (rv)  ||  bytes_read < 1)
				break;
			buf[bytes_read] = '\0';
			mPostData.Append (buf);
			read_total += bytes_read;
		}
		aStream->Close();
    }
    return NS_OK;
}


NS_IMETHODIMP nsCFHChannel::GetUploadStream (nsIInputStream **stream)
{
	NOT_IMPLEMENTED ("GetUpLoadStream");
}


//-----------------------------------------------------------------------------
// nsITransportEventSink
//-----------------------------------------------------------------------------

NS_IMETHODIMP nsCFHChannel::OnTransportStatus
	(nsITransport *trans, nsresult status,
     PRUint32 progress, PRUint32 progressMax)
{
    // suppress status notification if channel is no longer pending!
    if (mProgressSink && mRequest
			&& !(mLoadFlags & nsIFileChannel::LOAD_BACKGROUND)) {
        // file channel does not send OnStatus events!
        if (status == nsITransport::STATUS_READING ||
            status == nsITransport::STATUS_WRITING) {
            mProgressSink->OnProgress (NS_STATIC_CAST (nsIFileChannel*, this), nsnull, progress, progressMax);
        }
    }
    return NS_OK;
}


//-----------------------------------------------------------------------------
// nsIHttpChannel methods
//-----------------------------------------------------------------------------

NS_IMETHODIMP nsCFHChannel::GetResponseStatus (PRUint32 *aResponseStatus)
{
	*aResponseStatus = 200;	// OK
	return NS_OK;
}

NS_IMETHODIMP nsCFHChannel::GetReferrer (nsIURI * *aReferrer)
{
	NS_ENSURE_ARG_POINTER (aReferrer);
	*aReferrer = mReferrer;
	NS_IF_ADDREF (*aReferrer);
	return NS_OK;
}

NS_IMETHODIMP nsCFHChannel::SetReferrer (nsIURI * aReferrer)
{
	if (aReferrer) {
		nsIURI *clone;
		nsresult rv = aReferrer->Clone (&clone);
		if (NS_SUCCEEDED (rv))
			mReferrer = clone;
		return rv;
	}
	else {
		mReferrer = nsnull;
		return NS_OK;
	}
}


NS_IMETHODIMP nsCFHChannel::GetResponseHeader
(const nsACString & header, nsACString & _retval)
{
	return NS_OK;
}


NS_IMETHODIMP nsCFHChannel::SetResponseHeader
(const nsACString & header, const nsACString & value, PRBool merge)
{
	return NS_OK;
}


NS_IMETHODIMP nsCFHChannel::GetRequestHeader
(const nsACString & aHeader, nsACString & _retval)
{
	return NS_OK;
}


NS_IMETHODIMP nsCFHChannel::SetRequestHeader (const nsACString & aHeader, const nsACString & aValue, PRBool aMerge)
{
	return NS_OK;
}


NS_IMETHODIMP nsCFHChannel::GetRequestMethod (nsACString & aRequestMethod)
{ NOT_IMPLEMENTED ("GetRequestMethod"); }
NS_IMETHODIMP nsCFHChannel::SetRequestMethod (const nsACString & aRequestMethod)
{ NOT_IMPLEMENTED ("SetRequestMethod"); }
NS_IMETHODIMP nsCFHChannel::VisitRequestHeaders (nsIHttpHeaderVisitor *aVisitor)
{ NOT_IMPLEMENTED ("VisitRequestHeaders"); }
NS_IMETHODIMP nsCFHChannel::GetAllowPipelining (PRBool *aAllowPipelining)
{ NOT_IMPLEMENTED ("GetAllowPipelining"); }
NS_IMETHODIMP nsCFHChannel::SetAllowPipelining (PRBool aAllowPipelining)
{ NOT_IMPLEMENTED ("SetAllowPipelining"); }
NS_IMETHODIMP nsCFHChannel::GetRedirectionLimit (PRUint32 *aRedirectionLimit)
{ NOT_IMPLEMENTED ("GetRedirectionLimit"); }
NS_IMETHODIMP nsCFHChannel::SetRedirectionLimit (PRUint32 aRedirectionLimit)
{ NOT_IMPLEMENTED ("SetRedirectionLimit"); }
NS_IMETHODIMP nsCFHChannel::GetResponseStatusText (nsACString & aResponseStatusText)
{ NOT_IMPLEMENTED ("GetResponseStatusText"); }
NS_IMETHODIMP nsCFHChannel::GetRequestSucceeded (PRBool *aRequestSucceeded)
{ NOT_IMPLEMENTED ("GetRequestSucceeded"); }
NS_IMETHODIMP nsCFHChannel::VisitResponseHeaders (nsIHttpHeaderVisitor *aVisitor)
{ NOT_IMPLEMENTED ("VisitResponseHeaders"); }
NS_IMETHODIMP nsCFHChannel::IsNoStoreResponse (PRBool *_retval)
{ NOT_IMPLEMENTED ("IsNoStoreResponse"); }
NS_IMETHODIMP nsCFHChannel::IsNoCacheResponse (PRBool *_retval)
{ NOT_IMPLEMENTED ("IsNoCacheResponse"); }

#if defined (MOZILLA_1_5) || defined (MOZILLA_1_6)
NS_IMETHODIMP nsCFHChannel::GetRequestVersion(PRUint32 *major, PRUint32 *minor)
{ NOT_IMPLEMENTED ("GetRequestVersion"); }
NS_IMETHODIMP nsCFHChannel::GetResponseVersion(PRUint32 *major, PRUint32 *minor)
{ NOT_IMPLEMENTED ("GetResponseVersion"); }
NS_IMETHODIMP nsCFHChannel::SetCookie(const char *aCookieHeader)
{ NOT_IMPLEMENTED ("SetCookie"); } 
#endif

//-----------------------------------------------------------------------------
// nsIHttpChannelInternal methods
//-----------------------------------------------------------------------------

NS_IMETHODIMP nsCFHChannel::GetDocumentURI(nsIURI * *aDocumentURI)
{
	NS_ENSURE_ARG_POINTER(aDocumentURI);
	*aDocumentURI = mDocumentURI;
	NS_IF_ADDREF(*aDocumentURI);
	return NS_OK;
}


NS_IMETHODIMP nsCFHChannel::SetDocumentURI(nsIURI * aDocumentURI)
{
	mDocumentURI = aDocumentURI;
	return NS_OK;
}


//-----------------------------------------------------------------------------
// nsIPostData methods
//-----------------------------------------------------------------------------

NS_IMETHODIMP nsCFHChannel::GetPostData (char* *aPostData)
{
	*aPostData = (char*) mPostData.get();
	return NS_OK;
}

//-----------------------------------------------------------------------------
