/*
 * $Id: nsPythonToHTMLConv.cpp,v 1.10 2004/01/08 01:29:34 schilz Exp $
 *
 * Copyright (c) 2004  Dr. Thomas Schilz
 *
 * This file is part of MozPython.
 *
 * MozPython is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MozPython is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MozPython; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "Python.h"
#include "nsPythonToHTMLConv.h"
#include "nsNetUtil.h"
#include "nsIStringStream.h"
#include "nsIURLParser.h"
#include "nsIPostData.h"
#include "nsIFileChannel.h"
#include "nsIFile.h"
#include "nsIDocShell.h"
#include "nsDirectoryServiceDefs.h"
#include "nsIConsoleService.h"
#include <prlock.h>
#include <string>

#ifdef LINUX
#	include <dlfcn.h>
#endif

#define LOG(x)	printf ("%s: %s\n", __FILE__, x);

#define MIN(x,y) (((x) < (y)) ? (x) : (y))

#ifdef HAVE_WORKING_SETENV
#	ifdef WIN32
#		include <windows.h>
#		define setenv(n,v,o)	SetEnvironmentVariable (n, v)
#	endif
#endif

nsPythonToHTMLConv *ActiveConverter = NULL;

extern PRLock* PythonLock;


//---------------------------------------------------------------------------
// mozpython module
//---------------------------------------------------------------------------
static PyMethodDef mozpython_methods[] = {
	{NULL, NULL, 0, NULL}
};

static char *mozpython_doc = "\
This module provides access to the MozPython version and \
to the nsISupports interface of the request's target docshell.";

//---------------------------------------------------------------------------
// nsPythonToHTMLConv methods
//---------------------------------------------------------------------------

nsPythonToHTMLConv::nsPythonToHTMLConv() {
	NS_INIT_ISUPPORTS();
}


nsPythonToHTMLConv::~nsPythonToHTMLConv() {
}


NS_IMETHODIMP nsPythonToHTMLConv::SendData()
{
	nsCOMPtr<nsIInputStream> inputData;
	nsresult rv = NS_NewStringInputStream(getter_AddRefs(inputData), mBuffer);
	if (NS_FAILED(rv))
		return rv;

	return mListener->OnDataAvailable
						(mRequest, mContext, inputData, 0, mBuffer.Length());
}

//---------------------------------------------------------------------------

#ifndef HAVE_WORKING_SETENV
//
// When calling the library function setenv the environment variables
// defined here don't appear in Python's environment. Why???
//
// So we add the values directly as attributes to the os.environ dictionary.
//

void my_setenv (char *name, char *value)
{
	std::string stmt ("os.environ['");
	stmt += name;
	stmt += "']='";
	stmt += value;
	stmt += "'\n";
	PyRun_SimpleString (stmt.c_str());
}

#	define setenv(n,v,o) my_setenv (n, v);
#endif

void nsPythonToHTMLConv::SetupEnvironment()
{
#ifndef HAVE_WORKING_SETENV
	PyRun_SimpleString("import os\n");
#endif

	setenv ("SERVER_SOFTWARE", "MozPython", 1);
	setenv ("SERVER_ADMIN", "nobody", 1);
	setenv ("SERVER_PROTOCOL", "HTTP/1.1", 1);
	setenv ("HTTP_USER_AGENT", "Mozilla/5.0", 1);
	setenv ("HTTP_HOST", "nohost", 1);
	setenv ("SERVER_NAME", "", 1);
	setenv ("REMOTE_ADDR", "", 1);
	setenv ("REMOTE_PORT", "", 1);
	setenv ("SERVER_ADDR", "", 1);
	setenv ("SERVER_PORT", "", 1);
	setenv ("GATEWAY_INTERFACE", "CGI/1.1", 1);

	setenv ("REQUEST_URI", (char*) mPath.get(), 1);
    setenv ("SCRIPT_NAME", (char*) mFilePath.get(), 1);
    setenv ("SCRIPT_FILENAME", (char*) mFilePath.get(), 1);
    setenv ("PATH_TRANSLATED", (char*) mFilePath.get(), 1);

	setenv ("HTTP_KEEP_ALIVE", "300", 1);
	setenv ("HTTP_CONNECTION", "keep-alive", 1);
	setenv ("SHLVL", "0", 1);
	setenv ("HTTP_PRAGMA", "no-cache", 1);
	setenv ("HTTP_CACHE_CONTROL", "no-cache", 1);

	setenv ("HTTP_ACCEPT_CHARSET", "ISO-8859-1,utf-8;q=0.7,*;q=0.7", 1);
	setenv ("HTTP_ACCEPT", "text/xml,application/xml,application/xhtml+xml,text/html;q=0.9,text/plain;q=0.8,image/png,image/jpeg,image/gif;q=0.2,*/*;q=0.1", 1);
	setenv ("HTTP_ACCEPT_LANGUAGE", "en-us,en;q=0.5", 1);
	setenv ("HTTP_ACCEPT_ENCODING", "gzip,deflate ", 1);

    setenv ("DOCUMENT_ROOT", "/", 1);
	setenv ("QUERY_STRING", (char*) mQuery.get(), 1);

	if (mPostData.IsEmpty()) {
		setenv ("REQUEST_METHOD", "GET", 1);
	}
	else {
		char lbuf[16];
		sprintf (lbuf, "%ld", mPostData.Length());
		setenv ("REQUEST_METHOD", "POST", 1);
		setenv ("CONTENT_TYPE", "application/x-www-form-urlencoded", 1);
		setenv ("CONTENT_LENGTH", lbuf, 1);
	}

	/* Pass the script name as argv[0] to the Python interpreter. */
	Py_SetProgramName ((char*) mFilePath.get());
}



static nsresult GetDir (const char *key, nsCString &dir)
{
	dir.Assign (NS_LITERAL_CSTRING (""));

	nsCOMPtr<nsIServiceManager> servMan; 
	nsresult rv = NS_GetServiceManager(getter_AddRefs(servMan)); 

	if (NS_FAILED (rv)) return rv;

	nsCOMPtr<nsIProperties> directoryService; 
	rv = servMan->GetServiceByContractID (
			NS_DIRECTORY_SERVICE_CONTRACTID,  
			NS_GET_IID (nsIProperties), getter_AddRefs (directoryService)); 
 
	if (NS_FAILED (rv)) return rv;

	nsCOMPtr<nsIFile> theFile; 
	rv = directoryService->Get (key, NS_GET_IID (nsIFile),  
								getter_AddRefs (theFile)); 
 
	if (NS_FAILED (rv)) return rv;

	nsAutoString exe_dir;
	rv = theFile->GetPath (exe_dir);

	if (NS_FAILED (rv)) return rv;

	dir.AssignWithConversion (exe_dir);

	return NS_OK;
}



//---------------------------------------------------------------------------
// nsISupports methods
//---------------------------------------------------------------------------

NS_IMPL_THREADSAFE_ISUPPORTS3(nsPythonToHTMLConv,
							  nsIStreamConverter,
							  nsIRequestObserver,
							  nsIStreamListener);


//---------------------------------------------------------------------------
// nsIStreamConverter methods
//---------------------------------------------------------------------------

NS_IMETHODIMP nsPythonToHTMLConv::Convert
	(nsIInputStream *aFromStream, const PRUnichar *aFromType, const PRUnichar *aToType,
	 nsISupports *aCtxt, nsIInputStream * *_retval)
{
	return NS_ERROR_NOT_IMPLEMENTED;
}


NS_IMETHODIMP nsPythonToHTMLConv::AsyncConvertData
	(const PRUnichar *aFromType, const PRUnichar *aToType,
	 nsIStreamListener *aListener, nsISupports *aCtxt)
{
	NS_ASSERTION(aListener, "null pointer");
	mListener = aListener;
	return NS_OK;
}


//---------------------------------------------------------------------------
// nsIRequestObserver methods
//---------------------------------------------------------------------------

NS_IMETHODIMP nsPythonToHTMLConv::OnStartRequest
	(nsIRequest* request, nsISupports *aContext)
{
	mRequest = request;
	mContext = aContext;
	mBuffer.Assign (NS_LITERAL_STRING (""));

	nsCString mozBinDir;
	nsresult rv = GetDir (NS_XPCOM_CURRENT_PROCESS_DIR, mozBinDir);

	nsCString mozCompDir;
	rv = GetDir (NS_XPCOM_COMPONENT_DIR , mozCompDir);

	nsCOMPtr<nsIChannel> channel = do_QueryInterface (request);
	NS_ENSURE_TRUE (channel, NS_ERROR_UNEXPECTED);
	channel->SetContentType (NS_LITERAL_CSTRING ("text/html"));

	nsCOMPtr<nsIURI> uri;
	channel->GetURI (getter_AddRefs (uri));
	NS_ENSURE_TRUE(uri, NS_ERROR_UNEXPECTED);

	nsCOMPtr<nsIURLParser> url_parser = do_GetService (NS_STDURLPARSER_CONTRACTID);
	NS_ENSURE_TRUE (url_parser, NS_ERROR_UNEXPECTED);

	rv = uri->GetSpec (mURI);
	if (NS_FAILED(rv))
		return rv;

	PRUint32 schemePos, authPos, pathPos;
	PRInt32 schemeLen, authLen, pathLen;

	url_parser->ParseURL (mURI.get(), -1,
						  &schemePos, &schemeLen,
						  &authPos, &authLen,
						  &pathPos, &pathLen);

	if (pathPos < 0 || pathLen < 0) {
		NS_ERROR ("Unable to determine path");
		return NS_BASE_STREAM_ILLEGAL_ARGS;
	}

	mPath = nsCAutoString (mURI.get() + pathPos, pathLen);

	PRUint32 filePos, paramPos, queryPos, refPos;
	PRInt32 fileLen, paramLen, queryLen, refLen;

	url_parser->ParsePath (mPath.get(), pathLen,
						   &filePos, &fileLen,
						   &paramPos, &paramLen,
						   &queryPos, &queryLen,
						   &refPos, &refLen);

	if (queryLen > 0)
		mQuery.Assign (nsCAutoString (mPath.get() + queryPos, queryLen));
	else
		mQuery.Assign (NS_LITERAL_CSTRING (""));

	nsCOMPtr<nsIPostData> ipoda (do_QueryInterface(channel));
	if (ipoda) {
		char *p;
		if (NS_SUCCEEDED (ipoda->GetPostData (&p))) {
			nsCAutoString s (p);

			// Discard header and empty line separating it from body
			nsCAutoString::const_iterator start, end;
			s.BeginReading (start); s.EndReading (end);
			if (FindInReadable (nsCAutoString ("\r\n\r\n"), start, end))
				s.Right (mPostData, end.size_forward());
		}
	}


	/*
	 * Extract the path to the Python script file to be executed.
	 */
	nsCOMPtr<nsIFileChannel> file_channel = do_QueryInterface (request);
	NS_ENSURE_TRUE (file_channel, NS_ERROR_UNEXPECTED);

	nsCOMPtr<nsIFile> file;
	rv = file_channel->GetFile (getter_AddRefs (file));
	if (NS_FAILED(rv)) return rv;

	nsAutoString file_path;
	rv = file->GetPath (file_path);
	if (NS_FAILED(rv)) return rv;

	mFilePath.AssignWithConversion (file_path);

	char *script_path = (char*) mFilePath.get();

	FILE *script_fp = fopen (script_path, "r");

	if (script_fp == NULL) {
		NS_ERROR ("Open script failed.");
		return NS_ERROR_UNEXPECTED;
	}

	/*
	 * Now we can inform the listener about the request. We will at least
	 * send some error message, should the execution of the script fail.
	 */
	rv = mListener->OnStartRequest(request, aContext);
	if (NS_FAILED(rv))
		return rv;

	/*
	 * Lock the Python interpreter to ensure that no other thread uses
	 * it concurrently.
	 */
	PR_Lock (PythonLock);

	ActiveConverter = this;

#ifdef LINUX
	/*
	 * Work around for the following error:
	 *	ImportError: /usr/local/lib/python2.3/lib-dynload/cStringIO.so:
	 *	undefined symbol: PyObject_SelfIter
	 *
	 * This problem only occurs with Mozilla when configured with
	 *	-enable-optimze -disable-debug
	 *
	 * I suspect, that Mozilla then loads shared libraries without the
	 * RTLD_GLOBAL flag to dlopen. It does not happen with a debug version
	 * of Mozilla!? Anyway, reloading the Python shared library
	 * here solves the problem.
	 */
	void *lib_python = dlopen (LIB_PYTHON, RTLD_GLOBAL | RTLD_NOW);
#endif

	/*
	 * Initialize the Python interpreter if it has not yet been initialized.
	 * If pyxpcom is installed then Python may already be initialized.
	 */
	bool python_was_initialized = Py_IsInitialized();
	if (!python_was_initialized) {
		Py_Initialize();
		PyEval_InitThreads();
	}

	SetupEnvironment();

	PyObject *str_sys = PyString_FromString ("sys");
	PyObject *mod_sys = PyImport_Import (str_sys);
	Py_XDECREF (str_sys);

	PyObject *str_cStringIO = PyString_FromString ("cStringIO");
	PyObject *mod_cStringIO = PyImport_Import (str_cStringIO);
	Py_XDECREF (str_cStringIO);

	/*
	 * Redirect sys.stdout and sys.stderr to cStringIO instances.
	 */
	PyObject *out_StringIO = PyObject_CallMethod (mod_cStringIO, "StringIO", NULL);
	PyObject *err_StringIO = PyObject_CallMethod (mod_cStringIO, "StringIO", NULL);

	PyObject_SetAttrString (mod_sys, "stdout", out_StringIO);
	PyObject_SetAttrString (mod_sys, "stderr", err_StringIO);

	/*
	 * If there is POST data available feed it to sys.stdin by means of
	 * another cStringIO object.
	 */
	if (!mPostData.IsEmpty()) {
		PyObject *str_StringIO = PyString_FromString ("StringIO");
		PyObject *str_poda = PyString_FromString (mPostData.get());
		PyObject *in_StringIO = PyObject_CallMethodObjArgs (mod_cStringIO, str_StringIO, str_poda, NULL);
		Py_XDECREF (str_StringIO);
		Py_XDECREF (str_poda);
		PyObject_SetAttrString (mod_sys, "stdin", in_StringIO);
		Py_XDECREF (in_StringIO);
	}

	Py_XDECREF (mod_cStringIO);

	/*
	 * Create a "mozpython" module with two attributes:
	 * - version: The MozPython version
	 * - docShell: The nsISupport interface of the request's target nsDocShell.
	 *		This attribute is defined only if MozPython can import the PyXPCOM
	 *		module.
	 */
	PyObject *mod_mozpython =
			Py_InitModule3 ("mozpython", mozpython_methods, mozpython_doc);

	PyObject *str_MozPythonVersion = PyString_FromString (MOZPYTHON_VERSION);
	PyObject_SetAttrString (mod_mozpython, "version", str_MozPythonVersion);
	Py_XDECREF (str_MozPythonVersion);


	/*
	 * Prepend the PyXPCOM directory to sys.path if it's not yet already
	 * in there.
	 */
	PyObject *sys_path = PyObject_GetAttrString (mod_sys, "path");

	if (sys_path != NULL) {
		nsCString pyxpcom_path (mozBinDir);
		pyxpcom_path.Append ("/python");
		PyObject *ppObj = PyString_FromString (pyxpcom_path.get());
		if (ppObj != NULL) {
			if (!PySequence_Contains (sys_path, ppObj))
				PyList_Insert (sys_path, 0, ppObj);
			Py_XDECREF (ppObj);
		}
		Py_XDECREF (sys_path);
	}

	/*
	 * Import the xpcom module to get it initialized. We will call
	 * one of it's internal functions to create the docShell object.
	 */
	int pr = PyRun_SimpleString ("import xpcom._xpcom");

	if (pr == 0) {
		/*
	 	 * Try to load the mpdlpxpc library and resolve the functions we
		 * need for creating the docShell object of our mozpython module.
	 	 */ 

		nsCString libmpdlpxpc_name (mozBinDir);

#ifdef WIN32
# ifdef DEBUG
		libmpdlpxpc_name.Append ("/python/xpcom/mpdlpxpc_d.dll");
# else
		libmpdlpxpc_name.Append ("/python/xpcom/mpdlpxpc.dll");
# endif
#else
		libmpdlpxpc_name.Append ("/python/xpcom/libmpdlpxpc.so");
#endif

		PRLibrary *libmpdlpxpc = PR_LoadLibrary (libmpdlpxpc_name.get());

		if (libmpdlpxpc != NULL) {
			PyObject* (*PyObjectFromInterface)
								(nsISupports *pis, const nsIID &riid,
								 PRBool bAddRef,
								 PRBool bMakeNicePyObject = PR_TRUE) =
				(PyObject*(*)(nsISupports*, const nsIID&, PRBool, PRBool))
						PR_FindSymbol (libmpdlpxpc, "PyObjectFromInterface");

			/*
 			 * Get a pointer to the target nsDocShell of the request.
 			 * The nsDocShell is casted to nsIInterfaceRequestor*
 			 * and passed to the channel in nsDocShell::DoURILoad. See
 			 * the call to NS_NewChannel there.
 			 * NS_NewChannel then assigns the nsIInterfaceRequestor pointer
			 * to the channel's "callbacks" attribute.
 			 */

			PyObject *po_docShell = NULL;

			nsIInterfaceRequestor *ifReq;
			rv = channel->GetNotificationCallbacks (&ifReq);

			if (NS_SUCCEEDED (rv)  &&  PyObjectFromInterface) {
				const nsIID isIID = NS_ISUPPORTS_IID;
				nsCOMPtr<nsISupports> pis = do_QueryInterface (ifReq);
				po_docShell = PyObjectFromInterface (pis, isIID, PR_FALSE);
			}

			if (po_docShell)
				/*
				 * Assign the nsDocShell's nsISupports interface to an
				 * attribute named "docShell" of module mozpython.
				 */
				PyObject_SetAttrString (mod_mozpython, "docShell", po_docShell);
			else // something went wrong...
				PyObject_SetAttrString (mod_mozpython, "docShell", Py_None);

			(void) PR_UnloadLibrary (libmpdlpxpc);
		} // libmpdlpxpc != NULL
	} // pr == 0

	char *argv[] = { "mozilla" };
	PySys_SetArgv (1, argv);

	Py_XDECREF (mod_sys);

	/*
	 * Load and execute the script file.
	 */
	int py_result = PyRun_SimpleFile (script_fp, script_path);

	fclose (script_fp);

	PyObject *out_val = PyObject_CallMethod (out_StringIO, "getvalue", NULL);
	char *c_out = PyString_AsString (out_val);

	PyObject *err_val = PyObject_CallMethod (err_StringIO, "getvalue", NULL);
	char *c_err = PyString_AsString (err_val);

	mBuffer.AssignWithConversion (c_out);

	Py_XDECREF (out_val);

	(void) PyObject_CallMethod (out_StringIO, "close", NULL);
	Py_XDECREF (out_StringIO);

	nsCOMPtr<nsIConsoleService> aConsoleService = 
		do_GetService ("@mozilla.org/consoleservice;1");

	if (py_result != 0)
		aConsoleService->LogStringMessage (
			NS_LITERAL_STRING ("MozPython: Error processing Python script file").get());
	if (c_err != NULL  &&  *c_err != '\0') {
		nsString msg;
		msg.AssignWithConversion (c_err);
		aConsoleService->LogStringMessage (msg.get());
	}

	(void) PyObject_CallMethod (err_StringIO, "close", NULL);
	Py_XDECREF (err_StringIO);

#ifdef notdef
	// Finalizing the interpreter here makes subsequent imports of the
	// xpcom module fail.
	if (!python_was_initialized)
		Py_Finalize();
#endif

#ifdef LINUX
	if (lib_python != NULL)
		(void) dlclose (lib_python);
#endif

	/*
	 * Unlock the Python interpreter.
	 */

	ActiveConverter = NULL;

	if (PR_Unlock (PythonLock) != PR_SUCCESS)
		return NS_ERROR_UNEXPECTED;

	rv = SendData();

	return rv;
}


NS_IMETHODIMP nsPythonToHTMLConv::OnStopRequest
	(nsIRequest* request, nsISupports *aContext, nsresult aStatus)
{
	return mListener->OnStopRequest(request, aContext, aStatus);
}


//---------------------------------------------------------------------------
// nsIStreamListener method
//---------------------------------------------------------------------------

NS_IMETHODIMP nsPythonToHTMLConv::OnDataAvailable
	(nsIRequest* request, nsISupports *aContext,
	 nsIInputStream *aInStream, PRUint32 aOffset, PRUint32 aCount)
{
	/*
	 * Nothing to be done here.
	 * Input data is read by the Python interpreter in OnStartRequest.
	 *
	 * But we have to do dummy reads here to get the request finished.
	 *
	 * Note: We don't get here before the Python interpreter has completed
	 * reading and executing the script file.
	 */

	PRUint32 amtRead = 0;
	char *buffer = (char*) nsMemory::Alloc (aCount + 1);
	if (!buffer) return NS_ERROR_OUT_OF_MEMORY;
	
	while (amtRead < aCount) {
		PRUint32 read = 0;
		nsresult rv = aInStream->Read (buffer, aCount - amtRead, &read);
		if (NS_FAILED(rv)) return rv;
		amtRead += read;
	}

	return NS_OK; 
} 
